/*
 * DiffDialog.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Paint;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.RowSorter.SortKey;
import javax.swing.SortOrder;
import javax.swing.SwingWorker;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import jp.co.ntt.oss.heapstats.container.MapKey;
import jp.co.ntt.oss.heapstats.container.ObjectData;
import jp.co.ntt.oss.heapstats.container.PackageFilter;
import jp.co.ntt.oss.heapstats.container.SnapShot;
import jp.co.ntt.oss.heapstats.gui.graph.GraphColorHelper;
import jp.co.ntt.oss.heapstats.gui.snapshot.PackageFilterReader;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotHelper;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotReader;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.CategoryLabelPositions;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.data.category.DefaultCategoryDataset;

/**
 * extends {@link JDialog}<br>
 * implements {@link PropertyChangeListener}<br>
 * <br>
 * Delta Index chart, draw a stacked heap usage chart. In addition, articles can
 * be done by filter class.
 */
public class DiffDialog extends JDialog implements
                                       PropertyChangeListener,
                                       MouseListener {
    /** serialVersionUID. */
    private static final long serialVersionUID = 31766154511197856L;

    /** Angle of the axis labels. */
    private static final double ANGLE = 6.0;

    /** The button name for choosing all the classes for search. */
    private static final String SELECT_BUTTON_NAME = " Select All ";

    /** A SnapShot Dialog. */
    private SnapShotDialog ssDlg;

    /** A ReferenceTree Dialog. */
    private ReferenceDialog refDlg;

    /** Mainframe heapstats. */
    private GUIMain guimain;

    /** Main Component. */
    private JSplitPane comp;

    /** Delta data information display unit. */
    private Box deltaUnit = Box.createVerticalBox();

    /** Ranking Chart Panel. */
    private ExpansionContractionPanel rank;

    /** Heap Usage Chart Panel. */
    private ExpansionContractionPanel usage;

    /** Objects List. */
    private ExpansionContractionPanel objList;

    /** Objects List DefaultTableModel. */
    private DiffEntriesModel diffEntriesModel;

    /** Objects List Table. */
    private JTable diffEntries;

    /** Class Filter TableModel. */
    private FilterEntriesModel filterEntriesModel;

    /** Class Filter Table. */
    private JTable filterEntries;

    /** Class Filter OK Button. */
    private JButton addOkbtn;

    /** Class Filter Clean Button. */
    private JButton clearBtn;

    /** Class Filter Information. */
    private PackageFilter packageFilter;

    /** Regular expression. */
    private List<Pattern> filterList;

    /** Filter Results Class List. */
    private Object[] searchClass;

    /** Class Search List. */
    private JList classList;

    /** Class Search. */
    private SearchJTextField searchText;

    /** Class Search OK Button. */
    private JButton searchOkbtn;

    /** Class Search Clear Button. */
    private JButton searchClearbtn;

    /** Select button. */
    private JButton selectBtn;

    /** Sort Button. instance */
    private JButton insBtn;

    /** Sort Button. heap size */
    private JButton sizeBtn;

    /** A class list search paths. */
    private Set<String> nameList;

    /** Offsets Key is Date. Data is file position and SnapShotReader. */
    private Map<String, SnapShotReader> offsets;

    /** the List is file position and SnapShotReader. */
    private List<SnapShotReader> diffOffsets;

    /** ProgressBar. */
    private HeapProgressBar progressBar;

    /** The SwingWorker for drawing graphs. */
    private DiffWorker worker;

    /** The panel of Ranking Chart. */
    private ChartPanel rankChart;

    /** The panel of Usage Chart. */
    private ChartPanel usageChart;

    /**
     * Creates a modeless dialog with the specified Dialog as its owner.
     *
     * @param owner the owner Dialog from which the dialog is displayed or null
     *        if this dialog has no owner
     * @param snapshotDialog the snapshot dialog
     * @param referenceDialog the referencetree dialog
     */
    public DiffDialog(final GUIMain owner, final SnapShotDialog snapshotDialog,
            final ReferenceDialog referenceDialog) {
        super(owner);

        guimain = owner;
        ssDlg = snapshotDialog;
        refDlg = referenceDialog;
        rank = null;
        usage = null;
        objList = null;
        comp = null;
        searchClass = null;
        filterList = new ArrayList<Pattern>();
        packageFilter = new PackageFilter();
        nameList = new HashSet<String>();

        setTitle(GuiCommon.getResources().getString(GuiCommon.TAB_TITLE_DELA));
        setSize(GuiCommon.DEFAULT_DIALOG_WIDTH, GuiCommon.DEFAULT_DIALOG_HEIGHT);
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        setModal(false);

        addComponent();
        setLocationByPlatform(true);
    }

    /**
     * Create components to draw on the screen.
     */
    private void addComponent() {
        Box search = Box.createVerticalBox();
        search.setBorder(BorderFactory.createEmptyBorder(GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE));

        JPanel diffPanel = new JPanel();
        diffPanel.setLayout(new BorderLayout());
        deltaUnit.setBorder(BorderFactory.createEmptyBorder(
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE));

        comp = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, search, diffPanel);
        comp.setOneTouchExpandable(true);

        search.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
        search.add(addFilterComponent());
        search.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
        search.add(addIncrementComponent());

        diffPanel.add(deltaUnit);

        diffEntriesModel = new DiffEntriesModel();
        diffEntries = new JTable(diffEntriesModel) {
            /** serialVersionUID. */
            private static final long serialVersionUID = -925317579655004381L;

            @Override
            public Component prepareRenderer(final TableCellRenderer tcr,
                    final int row, final int column) {
                Component retVal = super.prepareRenderer(tcr, row, column);

                retVal.setForeground(getForeground());
                Paint bgPaint = GraphColorHelper.getColor((String) getValueAt(
                        row, 1));
                if (bgPaint != null && column == 0) {
                    retVal.setBackground((Color) bgPaint);
                } else {
                    if (isRowSelected(row)) {
                        retVal.setBackground(getSelectionBackground());
                    } else {
                        retVal.setBackground(Color.WHITE);
                    }
                }
                return retVal;
            }

            @Override
            public boolean isCellEditable(final int row, final int column) {
                return false;
            }
        };

        TableColumnModel model = diffEntries.getColumnModel();
        model.getColumn(0).setPreferredWidth(GuiCommon.COL_SIZE_COLOR);
        model.getColumn(1).setPreferredWidth(GuiCommon.COL_SIZE_CLASSNAME);
        diffEntries.removeColumn(diffEntries.getColumn(GuiCommon.getResources()
                .getString(GuiCommon.OBJECTLIST_CLASSTAG)));
        diffEntries.addMouseListener(this);

        rank = new ExpansionContractionPanel(GuiCommon.getResources()
                .getString(GuiCommon.TITLE_SIZE_DELTE));
        usage = new ExpansionContractionPanel(GuiCommon.getResources()
                .getString(GuiCommon.TITLE_SIZE_AREA));
        objList = new ExpansionContractionPanel(GuiCommon.getResources()
                .getString(GuiCommon.TITLE_OBJECTLIST));
    }

    /**
     * Creates a Class filter component.
     *
     * @return Return the Filter ExpansionContractionPanel
     */
    private JPanel addFilterComponent() {
        // Class Filter
        ExpansionContractionPanel filterMain = new ExpansionContractionPanel(
                GuiCommon.getResources().getString(
                        GuiCommon.CLASS_FILTER_LABEL_TITLE));
        JPanel classPanel = new JPanel(new BorderLayout());
        filterMain.setComponent(classPanel, true);
        classPanel.setBackground(Color.WHITE);

        JPanel addPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        addPanel.setBackground(Color.WHITE);

        JButton addFilter = new JButton(GuiCommon.getResources().getString(
                GuiCommon.CLASS_FILTER_ADD));
        addFilter.setPreferredSize(new Dimension(120, 20));
        addFilter.addActionListener(guimain);

        addPanel.add(addFilter);

        filterEntriesModel = new FilterEntriesModel();
        filterEntries = new JTable(filterEntriesModel);

        JScrollPane pkgScroll = new JScrollPane();
        pkgScroll.getViewport().setView(filterEntries);

        JPanel okPanel = new JPanel();
        okPanel.setLayout(new BoxLayout(okPanel, BoxLayout.X_AXIS));
        okPanel.setBackground(Color.WHITE);
        okPanel.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));

        clearBtn = new JButton("Clear");
        clearBtn.setMaximumSize(new Dimension(70, 20));
        clearBtn.addActionListener(guimain);
        okPanel.add(clearBtn);
        okPanel.add(new JLabel(" "));

        addOkbtn = new JButton("OK");
        addOkbtn.setMaximumSize(new Dimension(60, 20));
        addOkbtn.addActionListener(guimain);

        okPanel.add(addOkbtn);

        TableColumnModel model = filterEntries.getColumnModel();
        model.getColumn(0).setMaxWidth(50);

        classPanel.add(BorderLayout.NORTH, addPanel);
        classPanel.add(BorderLayout.CENTER, pkgScroll);
        classPanel.add(BorderLayout.SOUTH, okPanel);

        return filterMain;
    }

    /**
     * Creates a Class search component.
     *
     * @return Return the Incremental Search Component
     */
    private JPanel addIncrementComponent() {
        // Incremental Search
        ExpansionContractionPanel searchMain = new ExpansionContractionPanel(
                GuiCommon.getResources().getString(
                        GuiCommon.CLASS_SEARCH_LABEL_TITLE));
        JPanel increPanel = new JPanel(new BorderLayout());
        increPanel.setBackground(Color.WHITE);

        searchMain.setComponent(increPanel, true);
        JPanel okPanel = new JPanel();
        okPanel.setLayout(new BoxLayout(okPanel, BoxLayout.X_AXIS));
        okPanel.setBackground(Color.WHITE);

        searchClearbtn = new JButton("Clear");
        searchClearbtn.setMaximumSize(new Dimension(70, 20));
        searchClearbtn.addActionListener(guimain);
        searchOkbtn = new JButton("OK");
        searchOkbtn.setMaximumSize(new Dimension(60, 20));
        searchOkbtn.addActionListener(guimain);
        selectBtn = new JButton(SELECT_BUTTON_NAME);
        selectBtn.setMinimumSize(new Dimension(110, 20));
        selectBtn.setMaximumSize(new Dimension(110, 20));
        selectBtn.addActionListener(guimain);

        okPanel.add(selectBtn);
        okPanel.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
        okPanel.add(new JLabel(" "));
        okPanel.add(searchClearbtn);
        okPanel.add(new JLabel(" "));
        okPanel.add(searchOkbtn);

        DefaultListModel classModel = new DefaultListModel();
        classList = new JList(classModel);
        classList.setVisibleRowCount(50);
        JScrollPane listScroll = new JScrollPane();
        listScroll.getViewport().setView(classList);

        searchText = new SearchJTextField(0, classModel);

        // add(p1);
        increPanel.add(BorderLayout.NORTH, searchText);
        increPanel.add(BorderLayout.CENTER, listScroll);
        increPanel.add(BorderLayout.SOUTH, okPanel);

        return searchMain;
    }

    /**
     * Create an exclusion list for the class.
     */
    private void createFilterList() {
        // Class Filter
        filterList.clear();
        for (int i = 0; i < filterEntriesModel.getRowCount(); i++) {
            if ((Boolean) filterEntriesModel.getValueAt(i, 0)) {
                List<String> datas = packageFilter
                        .get((String) filterEntriesModel.getValueAt(i, 1));
                for (int j = 1; j < datas.size(); j++) {
                    StringBuilder buf = new StringBuilder();
                    buf.append("(^\\W*|L|\\W*L)?");
                    if (GuiCommon.getClassRename()) {
                        buf.append(datas.get(j).replace("\\", "."));
                    } else {
                        buf.append(datas.get(j));
                        if ((datas.get(j).lastIndexOf(";") == -1)
                                && (datas.get(j).lastIndexOf("*") == -1)) {
                            buf.append("(;)?");
                        }
                    }
                    filterList.add(Pattern.compile(buf.toString(),
                            Pattern.CASE_INSENSITIVE));
                }
            }
        }
    }

    /**
     * Filter class to return a list of articles ObjectData made.
     *
     * @param diff Class Map to be filtered
     * @return Return the class filtered list
     */
    private List<ObjectData> filterClassMap(final Map<Long, ObjectData> diff) {
        ObjectData[] list = new ObjectData[diff.size()];

        // Start Filter
        int i = 0;
        for (Map.Entry<Long, ObjectData> elmt : diff.entrySet()) {
            list[i++] = elmt.getValue();
        }

        return startFilter(list);
    }

    /**
     * Filter class to return a list of articles ObjectData made.
     *
     * @param list ObjectData
     * @return Return the ObjectData filtered list
     */
    private List<ObjectData> startFilter(final ObjectData[] list) {
        List<ObjectData> tmpList = new ArrayList<ObjectData>();
        List<ObjectData> viewSet = new ArrayList<ObjectData>();

        // Start Filter
        for (ObjectData elmt : list) {
            boolean match = false;
            for (Pattern ptn : filterList) {
                Matcher matcher = ptn.matcher(elmt.getName());
                if (matcher.matches()) {
                    match = true;
                    break;
                }
            }

            if (!match) {
                // Matches or, when there is no filter
                tmpList.add(elmt);
            }
        }

        // Class Search
        if (searchClass != null && searchClass.length > 0) {
            for (ObjectData data : tmpList) {
                for (Object path : searchClass) {
                    if (data.getName().equals(path)) {
                        viewSet.add(data);
                    }
                }
            }
        } else {
            // when there is no filter
            viewSet.addAll(tmpList);
        }
        return viewSet;
    }

    /**
     * Add a ExpansionContractionPanel component.
     *
     * @param component the ExpansionContractionPanel component
     */
    private void addComp(final Component component) {
        deltaUnit.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
        deltaUnit.add(component);
        deltaUnit.revalidate();
    }

    /**
     * Reload the "Delta Data" and "SnapShot Data".
     */
    private void reLoad() {
        if (diffOffsets != null && diffOffsets.size() > 0) {
            try {
                draw(diffOffsets, false);
                ssDlg.draw(diffOffsets.get(0), searchClass, filterList);
            } catch (IOException e) {
                GuiCommon.showMessage(this, GuiCommon.BASEE13,
                        JOptionPane.ERROR_MESSAGE);
            } catch (ClassNotFoundException e) {
                GuiCommon.showMessage(this, GuiCommon.BASEE13,
                        JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    /**
     * Draw a chart ranking chart and heap usage.
     *
     * @param sortItem Sort key
     */
    private void drawChart(final SnapShotHelper.SortItem sortItem) {
        progressBar = new HeapProgressBar("Draw the statistics GC.", guimain);
        progressBar.setVisible(true);

        if (sortItem == SnapShotHelper.SortItem.SIZE) {
            rank.setTitle(GuiCommon.getResources().getString(
                    GuiCommon.TITLE_SIZE_DELTE));
            usage.setTitle(GuiCommon.getResources().getString(
                    GuiCommon.TITLE_SIZE_AREA));
        } else {
            rank.setTitle(GuiCommon.getResources().getString(
                    GuiCommon.TITLE_INSTANCE_DELTE));
            usage.setTitle(GuiCommon.getResources().getString(
                    GuiCommon.TITLE_INSTANCE_AREA));
        }

        worker = new DiffWorker(sortItem, this);
        worker.addPropertyChangeListener(this);
        worker.execute();
    }

    /**
     * Returns the Regular expression.
     *
     * @return Returns the Regular expression.
     */
    public final List<Pattern> getFilterList() {
        return filterList;
    }

    /**
     * Returns the Filter Results Class List.
     *
     * @return Returns the Filter Results Class List.
     */
    public final Object[] getSearchClass() {
        return searchClass;
    }

    /**
     * Starts drawing the SnapShot Data.
     *
     * @param offsetsList file position and SnapShotReader
     * @param load load flag true:load, false:reload
     * @throws IOException If an I/O error has occurred.
     * @throws ClassNotFoundException Class of a serialized object cannot be
     *         found.
     */
    public final void draw(final List<SnapShotReader> offsetsList,
            final boolean load) throws IOException {
        deltaUnit.setVisible(false);
        deltaUnit.removeAll();
        nameList.clear();
        diffOffsets = offsetsList;

        if (load) {
            searchText.clear();
            searchClass = null;
            packageFilter.clear();
            filterEntriesModel.setRowCount(0);
        }

        // JTable:Objects List
        TableRowSorter<TableModel> sorter = new TableRowSorter<TableModel>(
                diffEntriesModel);
        sorter.setSortable(0, false);
        diffEntries.setRowSorter(sorter);
        diffEntries.getTableHeader().setReorderingAllowed(false);

        diffEntriesModel.setRowCount(0);

        SnapShot start = diffOffsets.get(0).getClassEntry();
        SnapShot end = diffOffsets.get(diffOffsets.size() - 1).getClassEntry();

        Map<Long, ObjectData> diff = SnapShotHelper.calculateDiff(
                start.getElements(), end.getElements());

        Object[] ary = new Object[6];
        List<ObjectData> objectList = filterClassMap(diff);

        for (ObjectData obd : objectList) {
            ary[0] = "";
            ary[1] = obd.getName();

            // ClassLoader
            if (obd.getClassLoaderTag() < 0) {
                ary[2] = "-";
            } else if (obd.getClassLoaderTag() == 0) {
                ary[2] = "<SystemClassLoader>";
            } else {
                ObjectData loader = diff.get(obd.getClassLoaderTag());
                if (loader == null) {
                    ary[2] = "-(0x" + Long.toHexString(obd.getClassLoader())
                            + ")";
                } else {
                    ary[2] = diff.get(obd.getClassLoaderTag()).getName()
                            + "(0x" + Long.toHexString(obd.getClassLoader())
                            + ")";
                }
            }
            ary[3] = obd.getCount();
            ary[4] = obd.getTotalSize();
            ary[5] = obd.getTag();

            diffEntriesModel.addRow(ary);
        }

        JPanel objPanel = new JPanel();
        objPanel.setLayout(new BorderLayout());
        objPanel.setBackground(Color.WHITE);

        JPanel sortPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        sortPanel.setBackground(Color.WHITE);
        insBtn = new JButton(GuiCommon.getResources().getString(
                GuiCommon.SORT_INSTANCE));
        sizeBtn = new JButton(GuiCommon.getResources().getString(
                GuiCommon.SORT_SIZE));
        insBtn.setPreferredSize(new Dimension(130, 20));
        sizeBtn.setPreferredSize(new Dimension(100, 20));
        insBtn.addActionListener(guimain);
        insBtn.setEnabled(true);
        sizeBtn.addActionListener(guimain);
        sizeBtn.setEnabled(false);
        sortPanel.add(insBtn);
        sortPanel.add(sizeBtn);

        JScrollPane diffScroll = new JScrollPane(
                JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        diffScroll.getViewport().setView(diffEntries);

        ArrayList<SortKey> sortKeys = new ArrayList<SortKey>();
        sortKeys.add(new SortKey(4, SortOrder.DESCENDING));
        diffEntries.getRowSorter().setSortKeys(sortKeys);

        objPanel.add(BorderLayout.NORTH, sortPanel);
        objPanel.add(BorderLayout.CENTER, diffScroll);
        objList.setComponent(objPanel, false);

        addComp(rank);
        addComp(usage);
        addComp(objList);

        drawChart(SnapShotHelper.SortItem.SIZE);
    }

    /**
     * Return the component.
     *
     * @return Return the Component
     */
    public final JSplitPane getComp() {
        return comp;
    }

    /**
     * Return the panel of Ranking chart.
     *
     * @return Return the panel of Ranking chart
     */
    public final ChartPanel getRankChart() {
        return rankChart;
    }

    /**
     * Return the panel of Usage chart.
     *
     * @return Return the panel of Usage chart
     */
    public final ChartPanel getUsageChart() {
        return usageChart;
    }

    /**
     * Return the offsets.
     *
     * @return Return the offsets
     */
    public final Map<String, SnapShotReader> getOffsets() {
        return offsets;
    }

    /**
     * Sets the dialog component.
     *
     * @param value the Main Component
     */
    public final void setComp(final JSplitPane value) {
        comp = value;
        add(comp);
    }

    /**
     * Clear the delta data.
     */
    public final void clear() {
        deltaUnit.setVisible(false);
        nameList.clear();
        if (diffOffsets != null) {
            diffOffsets.clear();
        }
        packageFilter.clear();
        filterList.clear();
        filterEntriesModel.setRowCount(0);
        searchText.clear();
        searchClass = null;
        selectBtn.setText(SELECT_BUTTON_NAME);
    }

    /**
     * Returns true if this task completed. Completion may be due to normal
     * termination, an exception, or cancellation -- in all of these cases, this
     * method will return true.
     *
     * @return true if this task completed
     */
    public final boolean isDone() {
        if (worker != null) {
            return worker.isDone();
        }
        return true;
    }

    /**
     * Invoked when an action occurs from GUIMain.
     *
     * @param arg0 ActionEvent
     */
    public final void actionPerformed(final ActionEvent arg0) {
        try {
            if (arg0.getActionCommand().equals(
                    GuiCommon.getResources().getString(
                            GuiCommon.CLASS_FILTER_ADD))) {

                JFileChooser fileChoser = new JFileChooser();
                fileChoser.setCurrentDirectory(GuiCommon.getCurrentDir());
                fileChoser.setMultiSelectionEnabled(true);
                fileChoser.setFileFilter(new FileNameExtensionFilter("XML",
                        "xml"));
                fileChoser.setAcceptAllFileFilterUsed(false);

                if (fileChoser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                    try {
                        GuiCommon.setCurrentDir(fileChoser
                                .getCurrentDirectory());
                    } catch (IOException e) {
                        GuiCommon.showMessage(this, GuiCommon.BASEE17,
                                JOptionPane.ERROR_MESSAGE);
                    }
                    PackageFilterReader reader = new PackageFilterReader(
                            fileChoser.getSelectedFile());

                    if (!packageFilter.add(reader.loadFilterFIle(this),
                            filterEntriesModel)) {
                        // Duplicate filter name
                        GuiCommon.showMessage(this, GuiCommon.BASEE10,
                                JOptionPane.ERROR_MESSAGE);
                    }
                }
            } else if (arg0.getSource().equals(addOkbtn)) {
                createFilterList();
                if (filterEntriesModel.getRowCount() > 0) {
                    GuiCommon.showMessage(this, GuiCommon.BASEI01,
                            JOptionPane.INFORMATION_MESSAGE);
                    reLoad();
                }
            } else if (arg0.getSource().equals(searchOkbtn)) {
                // Class Search
                Object[] tmpSelect = classList.getSelectedValues();
                if ((tmpSelect.length > 0)
                        || ((tmpSelect.length == 0) && (searchClass != null) && (searchClass.length > 0))) {
                    searchClass = tmpSelect;
                    GuiCommon.showMessage(this, GuiCommon.BASEI01,
                            JOptionPane.INFORMATION_MESSAGE);
                    reLoad();
                }
            } else if (arg0.getSource().equals(searchClearbtn)) {
                // Class Search Clear
                searchText.setText("");
                selectBtn.setText(SELECT_BUTTON_NAME);
                if (searchClass != null && searchClass.length > 0) {
                    searchClass = new Object[0];
                    reLoad();
                }
            } else if (arg0.getSource().equals(selectBtn)) {
                // Select class list.
                if (selectBtn.getText().equals(SELECT_BUTTON_NAME)
                        && (classList.getModel().getSize() > 0)) {
                    int[] indices = new int[classList.getModel().getSize()];
                    for (int i = 0; i < indices.length; i++) {
                        indices[i] = i;
                    }
                    classList.setSelectedIndices(indices);
                    selectBtn.setText("Deselect All");
                } else {
                    classList.clearSelection();
                    selectBtn.setText(SELECT_BUTTON_NAME);
                }
            } else if (arg0.getSource().equals(clearBtn)) {
                // Class Filter Clear
                filterEntriesModel.setRowCount(0);
                packageFilter.clear();

                if (filterList.size() > 0) {
                    filterList.clear();
                    reLoad();
                }
            } else if (arg0.getSource().equals(insBtn)) {
                // Instances Rank & Delta graphs
                drawChart(SnapShotHelper.SortItem.INSTANCE);
                insBtn.setEnabled(false);
                sizeBtn.setEnabled(true);
            } else if (arg0.getSource().equals(sizeBtn)) {
                // Size Rank & Delta graphs
                drawChart(SnapShotHelper.SortItem.SIZE);
                insBtn.setEnabled(true);
                sizeBtn.setEnabled(false);
            }
        } catch (Throwable e) {
            GuiCommon.showMessage(this, GuiCommon.BASEE13,
                    JOptionPane.ERROR_MESSAGE);
        }
    }

    @Override
    public final void propertyChange(final PropertyChangeEvent evt) {
        String strPropertyName = evt.getPropertyName();
        if ("progress".equals(strPropertyName)) {
            int progress = (Integer) evt.getNewValue();
            progressBar.setValue(progress);
        }
    }

    @Override
    public final void mouseClicked(final MouseEvent arg0) {
        Object eventSource = arg0.getSource();

        if (eventSource.equals(diffEntries)) {
            if ((arg0.getClickCount() == 2)
                    && (arg0.getButton() == MouseEvent.BUTTON1)) {
                //
                Point pt = arg0.getPoint();
                int idx = diffEntries.rowAtPoint(pt);
                if (idx < 0) {
                    return;
                }

                int row = diffEntries.convertRowIndexToModel(idx);
                refDlg.draw((Long) diffEntries.getModel().getValueAt(row, 5),
                        diffOffsets.get(diffOffsets.size() - 1));
            }
        }
    }

    @Override
    public void mouseEntered(final MouseEvent arg0) {
    }

    @Override
    public void mouseExited(final MouseEvent arg0) {
    }

    @Override
    public void mousePressed(final MouseEvent arg0) {
    }

    @Override
    public void mouseReleased(final MouseEvent arg0) {
    }

    /**
     * extends {@link SwingWorker}<br>
     * <br>
     * SwingWorker thread drawing done in the time-consuming process chart.
     */
    class DiffWorker extends SwingWorker<Object, Object> {

        /** Sort key object list. */
        private SnapShotHelper.SortItem sortItem;

        /** the owner DiffDialog. */
        private DiffDialog diffDialog;

        /**
         * Create a SwingWorker.
         *
         * @param sort sort item
         * @param parent the owner Dialog
         */
        public DiffWorker(final SnapShotHelper.SortItem sort,
                final DiffDialog parent) {
            sortItem = sort;
            diffDialog = parent;
        }

        @Override
        protected Object doInBackground() throws Exception {
            drawRankingChart();
            drawUsageChart();
            return null;
        }

        @Override
        protected void done() {
            super.done();
            progressBar.setVisible(false);
            usage.validate();

            deltaUnit.setVisible(true);
            diffDialog.validate();
            diffDialog.getOwner().setCursor(Cursor.getDefaultCursor());
        }

        /**
         * Draw a chart ranking the difference between snapshots.
         *
         * @throws IOException If an I/O error has occurred.
         * @throws ClassNotFoundException Class of a serialized object cannot be
         *         found.
         */
        private void drawRankingChart() throws IOException,
                ClassNotFoundException {
            SnapShot before;
            SnapShot after;
            Map<Long, ObjectData> diff;
            ObjectData[] tmpList;
            StringBuilder labelStr = new StringBuilder();

            DefaultCategoryDataset rankDataSet = new DefaultCategoryDataset();

            SimpleDateFormat sdf = new SimpleDateFormat(
                    GuiCommon.CHART_DATE_FORMAT);

            before = diffOffsets.get(0).getSnapShot();

            List<String> hideLabel = new ArrayList<String>();
            int k = diffOffsets.size() / 10;
            int size = diffOffsets.size();
            for (int i = 0; i < diffOffsets.size() - 1; i++) {
                after = diffOffsets.get(i + 1).getSnapShot();

                // x-axis labels
                labelStr.delete(0, labelStr.length());
                labelStr.append(sdf
                        .format(before.getHeader().getSnapShotDate()));

                diff = SnapShotHelper.calculateDiff(before.getElements(),
                        after.getElements());
                tmpList = SnapShotHelper.getSortedArray(diff, sortItem,
                        SnapShotHelper.SortOrder.DESC);

                List<ObjectData> result = startFilter(tmpList);

                int j = 1;
                for (ObjectData objectdata : result) {
                    rankDataSet.addValue(j, new MapKey(objectdata),
                            labelStr.toString());
                    if (j++ == GuiCommon.getRankLevel()) {
                        break;
                    }
                }

                if (k > 0 && (i % k != 0)
                        && !hideLabel.contains(labelStr.toString())) {
                    hideLabel.add(labelStr.toString());
                }
                setProgress(50 * i / size);
                before = after;
            }

            JFreeChart chart = ChartFactory.createLineChart("", "", "Rank",
                    rankDataSet, PlotOrientation.VERTICAL, false, true, false);
            chart.setBackgroundPaint(Color.WHITE);

            CategoryPlot plot = (CategoryPlot) chart.getPlot();
            plot.setBackgroundPaint(new Color(GuiCommon.getBgColor()));
            ((LineAndShapeRenderer) plot.getRenderer())
                    .setBaseShapesVisible(true);
            ValueAxis axis = plot.getRangeAxis();
            axis.setStandardTickUnits(NumberAxis.createIntegerTickUnits());
            axis.setInverted(true);
            axis.setAutoRange(false);
            axis.setRange(0.8, GuiCommon.getRankLevel() + 0.2);
            GraphColorHelper.adjustColor(rankDataSet, plot);

            hideCategoryLabel(plot.getDomainAxis(), hideLabel);

            rankChart = new ChartPanel(chart);

            rank.setComponent(rankChart, false);
            rankChart.addChartMouseListener(guimain);
        }

        /**
         * Draw a stacked heap usage.
         *
         * @throws IOException If an I/O error has occurred.
         * @throws ClassNotFoundException Class of a serialized object cannot be
         *         found.
         */
        private void drawUsageChart() throws IOException,
                ClassNotFoundException {
            long heapUsage;
            ObjectData[] result;
            StringBuilder labelStr = new StringBuilder();
            DefaultCategoryDataset stackDataSet = new DefaultCategoryDataset();

            SimpleDateFormat sdf = new SimpleDateFormat(
                    GuiCommon.CHART_DATE_FORMAT);

            offsets = new HashMap<String, SnapShotReader>();

            List<String> hideLabel = new ArrayList<String>();
            int size = diffOffsets.size();
            int k = size / 10;

            // SnapShot
            int per = 0;
            int rmv = 0;
            for (SnapShotReader reader : diffOffsets) {
                SnapShot snapShot = reader.getSnapShot();
                heapUsage = snapShot.getHeapUsage();

                String before = labelStr.toString();
                labelStr.delete(0, labelStr.length());
                labelStr.append(sdf.format(snapShot.getHeader()
                        .getSnapShotDate()));

                result = SnapShotHelper.getSortedArray(snapShot.getElements(),
                        sortItem, SnapShotHelper.SortOrder.DESC);

                List<ObjectData> viewList = startFilter(result);

                // SnapShot Data
                offsets.put(sdf.format(snapShot.getHeader().getSnapShotDate()),
                        reader);

                // Object
                int j = 0;
                for (ObjectData objectdata : viewList) {
                    if (j++ < GuiCommon.getRankLevel()) {
                        stackDataSet.addValue(
                                (double) objectdata.getTotalSize()
                                        / GuiCommon.BINARY_PREFIX, new MapKey(
                                        objectdata), labelStr.toString());
                        heapUsage -= objectdata.getTotalSize();
                    }
                    nameList.add(objectdata.getName());
                }

                if ((searchClass == null || searchClass.length == 0)
                        && (filterList == null || filterList.size() == 0)) {
                    stackDataSet.addValue((double) heapUsage
                            / GuiCommon.BINARY_PREFIX, new MapKey(0, "Others"),
                            labelStr.toString());
                }

                if (!before.equals(labelStr.toString()) && (k > 0)
                        && (rmv++ % k != 0)) {
                    hideLabel.add(labelStr.toString());
                }
                // Sets the progress bound property.
                setProgress(50 + 50 * (++per) / size);
            }

            searchText.addList(nameList);

            JFreeChart chart = ChartFactory.createStackedAreaChart("", "",
                    "Usage(KB)", stackDataSet, PlotOrientation.VERTICAL, false,
                    true, true);
            chart.setBackgroundPaint(Color.WHITE);

            CategoryPlot plot = chart.getCategoryPlot();
            plot.setBackgroundPaint(new Color(GuiCommon.getBgColor()));
            hideCategoryLabel(plot.getDomainAxis(), hideLabel);

            GraphColorHelper.adjustColor(stackDataSet, chart.getPlot());
            usageChart = new ChartPanel(chart);

            usage.setComponent(usageChart, true);

            usageChart.addChartMouseListener(guimain);
        }

        /**
         * Hide unwanted items out of the X axis label.
         *
         * @param ca CategoryAxis
         * @param hideLabel Hide item list
         */
        private void hideCategoryLabel(final CategoryAxis ca,
                final List<String> hideLabel) {
            ca.setAxisLineVisible(false);
            ca.setCategoryLabelPositions(CategoryLabelPositions
                    .createUpRotationLabelPositions(Math.PI / ANGLE));

            for (String label : hideLabel) {
                ca.setTickLabelPaint(label, new Color(0, 0, 0, 0));
            }
        }
    }
}
