/*
 * SnapShotHelper.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui.snapshot;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

import jp.co.ntt.oss.heapstats.container.ObjectData;

/**
 * Sort by heap usage or the number of instances the object.
 */
public final class SnapShotHelper {
    /** Sort key object list. */
    public enum SortItem {
        /** Sort the key number of instances. */
        INSTANCE,
        /** Sort key heap usage. */
        SIZE
    };

    /** Sort order object list. */
    public enum SortOrder {
        /** Ascending order. */
        ASC,
        /** Descending order. */
        DESC
    };

    /**
     * Don't let anyone instantiate this class.
     */
    private SnapShotHelper() {
    }

    /**
     * Heap usage to calculate the difference in the number of instances of the
     * specified object.
     *
     * @param before Snapshot was taken when the object list
     * @param after Snapshot was taken when the object list
     * @return Return the Result of the difference calculation
     */
    public static Map<Long, ObjectData> calculateDiff(
            final Map<Long, ObjectData> before,
            final Map<Long, ObjectData> after) {
        ObjectData t1Value, diffData;
        Map<Long, ObjectData> result = new HashMap<Long, ObjectData>();

        for (Map.Entry<Long, ObjectData> entry : after.entrySet()) {
            t1Value = before.get(entry.getKey());

            if (t1Value == null) {
                diffData = entry.getValue().clone();
            } else {
                diffData = new ObjectData();
                diffData.setTag(t1Value.getTag());
                diffData.setName(t1Value.getName());
                diffData.setClassLoader(t1Value.getClassLoader());
                diffData.setClassLoaderTag(t1Value.getClassLoaderTag());
                diffData.setCount(entry.getValue().getCount()
                        - t1Value.getCount());
                diffData.setTotalSize(entry.getValue().getTotalSize()
                        - t1Value.getTotalSize());
            }
            result.put(diffData.getTag(), diffData);
        }
        return result;
    }

    /**
     * Ascending to a key column or a specified list of objects, sorted in
     * descending order.
     *
     * @param data List of objects to be sorted
     * @param sortItem Column to sort
     * @param order Sorting Information
     * @return Return the Sort objects made
     */
    public static ObjectData[] getSortedArray(final Map<Long, ObjectData> data,
            final SortItem sortItem, final SortOrder order) {
        ObjectData[] result = getArray(data);
        if (sortItem == SortItem.INSTANCE) {
            Arrays.sort(result, new InstanceComparator(order));
        } else {
            Arrays.sort(result, new TotalSizeComparator(order));
        }

        return result;
    }

    /**
     * Return the ObjectData array.
     *
     * @param data Snapshot was taken when the object list
     * @return Return the ObjectData array
     */
    private static ObjectData[] getArray(final Map<Long, ObjectData> data) {
        return data.values().toArray(new ObjectData[0]);
    }

    /**
     * implements {@link Comparator}, {@link Serializable}<br>
     * <br>
     * Sort by comparing the number of instances of objects.
     */
    private static class InstanceComparator implements
                                           Comparator<ObjectData>,
                                           Serializable {
        /** serialVersionUID. */
        private static final long serialVersionUID = -568089295630679219L;

        /** Sorting. */
        private SortOrder order;

        /**
         * Constructor sets the sort information.
         *
         * @param value SortOrder
         */
        public InstanceComparator(final SortOrder value) {
            order = value;
        }

        /**
         * Compare the number of instances of two objects and returns the
         * result.
         *
         * @param o1 Object to compare
         * @param o2 Object to compare
         * @return Return the Comparison result
         */
        public int compare(final ObjectData o1, final ObjectData o2) {
            long diff;

            if (order == SortOrder.ASC) {
                diff = o1.getCount() - o2.getCount();
            } else {
                diff = o2.getCount() - o1.getCount();
            }

            if (diff == 0) {
                return 0;
            } else if (diff < 0) {
                return -1;
            } else {
                return 1;
            }
        }
    }

    /**
     * implements {@link Comparator}, {@link Serializable}<br>
     * <br>
     * Sort by comparing the amount of heap objects.
     */
    private static class TotalSizeComparator implements
                                            Comparator<ObjectData>,
                                            Serializable {
        /** serialVersionUID. */
        private static final long serialVersionUID = 6168971781062537310L;

        /** Sorting. */
        private SortOrder order;

        /**
         * Constructor sets the sort information.
         *
         * @param value SortOrder
         */
        public TotalSizeComparator(final SortOrder value) {
            order = value;
        }

        @Override
        public int compare(final ObjectData o1, final ObjectData o2) {
            long diff;

            if (order == SortOrder.ASC) {
                diff = o1.getTotalSize() - o2.getTotalSize();
            } else {
                diff = o2.getTotalSize() - o1.getTotalSize();
            }

            if (diff == 0) {
                return 0;
            } else if (diff < 0) {
                return -1;
            } else {
                return 1;
            }
        }
    }
}
